/***************************************************************************
                          App.cpp  -  the main game loop.
                             -------------------
    revisions

    begin                : Sun Dec 10 2000
    copyright            : (C) 2000 by Ryan Phillips
    email                : ryan@trolocsis.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "GFX.h"
#include "App.h"
#include "bool.h"
#include "Image.h"
#include "TileMap.h"

int App::Run( char *argv[], SDL_Surface* scr )
{
   int done=FALSE;
   level = 1;
   pushes = 0;

   screen = scr;   

   ImageStore.AddBitmap( IMAGE_BG,       "bmp/circlecheckered.png");
   ImageStore.AddBitmap( IMAGE_PLAYER,   "bmp/player.png");
   ImageStore.AddBitmap( IMAGE_SPACE,    "bmp/space.png");
   ImageStore.AddBitmap( IMAGE_BOX,      "bmp/box.png");
   ImageStore.AddBitmap( IMAGE_BOXFILLED,"bmp/box2.png");
   ImageStore.AddBitmap( IMAGE_WALL,     "bmp/wall.png");
   ImageStore.AddBitmap( IMAGE_CROSS,    "bmp/cross.png"); 
   
   fntMain = TTF_OpenFont("font/bankgothic.ttf", 12);
   if( fntMain == NULL )
      return FALSE;

   TTF_SetFontStyle(fntMain, 1);

   MapStore.ReadMap( level );
   changed_map = TRUE;
   SetDirty(TRUE);
   MapLengthX = MapStore.GetLimitX();
   MapLengthY = MapStore.GetLimitY();
   
   board  = SDL_CreateRGBSurface (screen->flags, MapLengthX*20, MapLengthY*20, bpp, 0, 0, 0, 0);       
   status = SDL_CreateRGBSurface (screen->flags, width, 50, bpp, 0, 0, 0, 0);          

   Draw();

   while( !done )
   {                        
      if( IsKeyDown( KEY_ESC ) )
         done = TRUE;

      if( IsKeyDown( KEY_UP ) )
      {
         if( IsLegalMove( UP ) )
         {
            SetDirty( TRUE );
            pushes++;
            Draw();            
         }
         while( IsKeyDown( KEY_UP ) ){
            poll_keyboard();
         }         
      }
      else if( IsKeyDown( KEY_LEFT ) )
      {
         if( IsLegalMove( LEFT ) )
         {
            SetDirty( TRUE );
            pushes++;
            Draw();
         }
         while( IsKeyDown( KEY_LEFT ) ){ 
            poll_keyboard(); 
         }
      }
      else if( IsKeyDown( KEY_RIGHT ) )
      {
         if( IsLegalMove( RIGHT ) )
         {
            SetDirty( TRUE );
            pushes++;
            Draw();
         }
         while( IsKeyDown( KEY_RIGHT ) ){ 
            poll_keyboard(); 
         }
      }
      else if( IsKeyDown( KEY_DOWN ) )
      {
         if( IsLegalMove( DOWN ) )
         {
            SetDirty( TRUE );
            pushes++;
            Draw();
         }
         while( IsKeyDown( KEY_DOWN ) ){ 
            poll_keyboard(); 
         }
      }
      else if( IsKeyDown( KEY_F1 ) )
      {
         ResetGame();
         pushes = 0;
         while( IsKeyDown( KEY_F1 ) ){ 
            poll_keyboard(); 
         }
      }

      if( MapStore.IsSolved() )
      {
            SetDirty( TRUE );
            MapStore.ReadMap( ++level );
            changed_map = TRUE;

            MapLengthX = MapStore.GetLimitX();
            MapLengthY = MapStore.GetLimitY();
            SDL_FreeSurface( board );

            board = SDL_CreateRGBSurface (screen->flags, MapLengthX*20, MapLengthY*20, bpp, 0, 0, 0, 0);    
            pushes = 0;

            Draw();
      }

      poll_keyboard();                 
   }

   SDL_FreeSurface( board );
   SDL_FreeSurface( status );

   return TRUE;
}

int App::IsLegalMove( int direction )
{   
   int PlayerX = MapStore.GetPlayerX();
   int PlayerY = MapStore.GetPlayerY();
   char Tile1, Tile2;
   static char prev = SPACE;

   switch( direction )
   {
   case UP:
      if( PlayerY-1 == 0 )
         return FALSE;
      
      Tile1 = MapStore.GetXY( PlayerX, PlayerY-1 );
      Tile2 = MapStore.GetXY( PlayerX, PlayerY-2 );

      if( Tile1 == WALL )
         return FALSE;    

      if( (Tile1 == BOX || Tile1 == BOXFILLED) && (Tile2 == BOX || Tile2 == BOXFILLED) )  // Check Double Boxes
         return FALSE;

      if( Tile1 == SPACE || Tile1 == CROSS)
      {
         MapStore.SetBox( PLAYER, PlayerX, PlayerY-1 );
         MapStore.SetBox( prev, PlayerX, PlayerY );
         prev = Tile1;
         return TRUE;
      }

      if( Tile1 == BOX )
      {
         if( Tile2 == CROSS )
         {            
            MapStore.SetBox( PLAYER, PlayerX, PlayerY-1 );
            MapStore.SetBox( BOXFILLED, PlayerX, PlayerY-2 );
            MapStore.SetBox( SPACE, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {            
            MapStore.SetBox( PLAYER, PlayerX, PlayerY-1 );
            MapStore.SetBox( BOX, PlayerX, PlayerY-2 );
            MapStore.SetBox( prev, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == WALL )
            return FALSE;
      }

      if( Tile1 == BOXFILLED )
      {
         if( Tile2 == CROSS )
         {                     
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX, PlayerY-1 );
            MapStore.SetBox( BOXFILLED, PlayerX, PlayerY-2 );            
            prev = CROSS;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {
            char c = MapStore.GetXY( PlayerX, PlayerY-1 );
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX, PlayerY-1 );
            MapStore.SetBox( BOX, PlayerX, PlayerY-2 );
            
            if( c == BOXFILLED )
               prev = CROSS;
            else
               prev = SPACE;
            return TRUE;
         }         
         if( Tile2 == WALL )
            return FALSE;
      }
            
      break;
   
   case DOWN:
      if( PlayerY+1 == MapLengthY )
         return FALSE;
      
      Tile1 = MapStore.GetXY( PlayerX, PlayerY+1 );
      Tile2 = MapStore.GetXY( PlayerX, PlayerY+2 );

      if( Tile1 == WALL )
         return FALSE;

      if( (Tile1 == BOX || Tile1 == BOXFILLED) && (Tile2 == BOX || Tile2 == BOXFILLED) )  // Check Double Boxes
         return FALSE;

      if( Tile1 == SPACE || Tile1 == CROSS)
      {
         MapStore.SetBox( PLAYER, PlayerX, PlayerY+1 );
         MapStore.SetBox( prev, PlayerX, PlayerY );
         prev = Tile1;
         return TRUE;
      }

      if( Tile1 == BOX )
      {
         if( Tile2 == CROSS )
         {            
            MapStore.SetBox( PLAYER, PlayerX, PlayerY+1 );
            MapStore.SetBox( BOXFILLED, PlayerX, PlayerY+2 );
            MapStore.SetBox( SPACE, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {            
            MapStore.SetBox( PLAYER, PlayerX, PlayerY+1 );
            MapStore.SetBox( BOX, PlayerX, PlayerY+2 );
            MapStore.SetBox( prev, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == WALL )
            return FALSE;
      }

      if( Tile1 == BOXFILLED )
      {
         if( Tile2 == CROSS )
         {                     
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX, PlayerY+1 );
            MapStore.SetBox( BOXFILLED, PlayerX, PlayerY+2 );            
            prev = CROSS;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {
            char c = MapStore.GetXY( PlayerX, PlayerY+1 );
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX, PlayerY+1 );
            MapStore.SetBox( BOX, PlayerX, PlayerY+2 );
            
            if( c == BOXFILLED )
               prev = CROSS;
            else
               prev = SPACE;
            return TRUE;
         }         
         if( Tile2 == WALL )
            return FALSE;
      }
   case RIGHT:
      if( PlayerX+1 == MapLengthX )
         return FALSE;
      
      Tile1 = MapStore.GetXY( PlayerX+1, PlayerY );
      Tile2 = MapStore.GetXY( PlayerX+2, PlayerY );

      if( Tile1 == WALL )
         return FALSE;      

      if( (Tile1 == BOX || Tile1 == BOXFILLED) && (Tile2 == BOX || Tile2 == BOXFILLED) )  // Check Double Boxes
         return FALSE;

      if( Tile1 == SPACE || Tile1 == CROSS)
      {
         MapStore.SetBox( PLAYER, PlayerX+1, PlayerY );
         MapStore.SetBox( prev, PlayerX, PlayerY );
         prev = Tile1;
         return TRUE;
      }

      if( Tile1 == BOX )
      {
         if( Tile2 == CROSS )
         {            
            MapStore.SetBox( PLAYER, PlayerX+1, PlayerY );
            MapStore.SetBox( BOXFILLED, PlayerX+2, PlayerY );
            MapStore.SetBox( SPACE, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {            
            MapStore.SetBox( PLAYER, PlayerX+1, PlayerY );
            MapStore.SetBox( BOX, PlayerX+2, PlayerY );
            MapStore.SetBox( prev, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == WALL )
            return FALSE;
      }

      if( Tile1 == BOXFILLED )
      {
         if( Tile2 == CROSS )
         {                     
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX+1, PlayerY );
            MapStore.SetBox( BOXFILLED, PlayerX+2, PlayerY );            
            prev = CROSS;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {
            char c = MapStore.GetXY( PlayerX+1, PlayerY );
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX+1, PlayerY );
            MapStore.SetBox( BOX, PlayerX+2, PlayerY );
            
            if( c == BOXFILLED )
               prev = CROSS;
            else
               prev = SPACE;
            return TRUE;
         }         
         if( Tile2 == WALL )
            return FALSE;
      }

      case LEFT:
         if( PlayerX-1 == MapLengthX )
            return FALSE;
      
      Tile1 = MapStore.GetXY( PlayerX-1, PlayerY );
      Tile2 = MapStore.GetXY( PlayerX-2, PlayerY );

      if( Tile1 == WALL )
         return FALSE;      

      if( (Tile1 == BOX || Tile1 == BOXFILLED) && (Tile2 == BOX || Tile2 == BOXFILLED) )  // Check Double Boxes
         return FALSE;

      if( Tile1 == SPACE || Tile1 == CROSS)
      {
         MapStore.SetBox( PLAYER, PlayerX-1, PlayerY );
         MapStore.SetBox( prev, PlayerX, PlayerY );
         prev = Tile1;
         return TRUE;
      }

      if( Tile1 == BOX )
      {
         if( Tile2 == CROSS )
         {            
            MapStore.SetBox( PLAYER, PlayerX-1, PlayerY );
            MapStore.SetBox( BOXFILLED, PlayerX-2, PlayerY );
            MapStore.SetBox( SPACE, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {            
            MapStore.SetBox( PLAYER, PlayerX-1, PlayerY );
            MapStore.SetBox( BOX, PlayerX-2, PlayerY );
            MapStore.SetBox( prev, PlayerX, PlayerY );            
            prev = SPACE;
            return TRUE;
         }
         if( Tile2 == WALL )
            return FALSE;
      }

      if( Tile1 == BOXFILLED )
      {
         if( Tile2 == CROSS )
         {                     
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX-1, PlayerY );
            MapStore.SetBox( BOXFILLED, PlayerX-2, PlayerY );            
            prev = CROSS;
            return TRUE;
         }
         if( Tile2 == SPACE )
         {
            char c = MapStore.GetXY( PlayerX-1, PlayerY );
            MapStore.SetBox( prev, PlayerX, PlayerY );
            MapStore.SetBox( PLAYER, PlayerX-1, PlayerY );
            MapStore.SetBox( BOX, PlayerX-2, PlayerY );
            
            if( c == BOXFILLED )
               prev = CROSS;
            else
               prev = SPACE;
            return TRUE;
         }         
         if( Tile2 == WALL )
            return FALSE;
      }


   default:
      break;
   }
   return FALSE;
}

int App::Draw()
{
   SDL_Rect dest;
   dest.x = (width-MapLengthX*20) >> 1;
   dest.y = (height-MapLengthY*20) >> 1;
   dest.w = MapLengthX*20;
   dest.h = MapLengthY*20;   

   if( IsDirty() )  
   {  
      if( changed_map )
      {
        DrawBG( TRUE );        
      }
      
      DrawText();
      DrawMap();
      DrawBorder();         

      if( changed_map )
          SDL_UpdateRect( screen, 0,0,0,0 );
      else
          SDL_UpdateRect( screen, dest.x,dest.y,dest.w,dest.h );
      SetDirty( FALSE );
      changed_map = false;
   }
   return TRUE;
}

int App::DrawBG( int Tiled ) // Centered?
{
   SDL_Surface* bg = ImageStore.Images[ IMAGE_BG ];
   int h = bg->h;
   int w = bg->w;
   SDL_Rect dest;

   dest.w = w;
   dest.h = h;
   dest.x = 0;
   dest.y = 0;

   if( Tiled )
   {
      for( int i=0; i<=height +  h; i+= h )
         for( int j=0; j<=width + w; j+= w )
         {
            SDL_BlitSurface( bg, 0, screen, &dest );
            dest.x = j;
            dest.y = i;
            dest.w = w; 
            dest.h = h;
         }
   }
   else  // Centered
   {
      dest.x = (width - w) / 2;
      dest.y = (height - h) / 2;
      SDL_BlitSurface( bg, 0, screen, &dest );
   }
   
   return TRUE;
}

int App::DrawMap()
{
   int y,x;
   SDL_Rect dest;   
   SDL_Surface *bitmap;    

   for (y = 0; y < MapLengthY; y++)
   {
      for (x = 0; x < MapLengthX; x++)
      {
         dest.x = (x*20);
         dest.y = (y*20);

         switch( MapStore.GetXY(x,y) )
         {                    
         case SPACE:
            bitmap = Image::Images[ IMAGE_SPACE ];
            dest.w = bitmap->w;
            dest.h = bitmap->h;            
            break;
         case WALL:
            bitmap = Image::Images[ IMAGE_WALL ];
            dest.w = bitmap->w;
            dest.h = bitmap->h;            
            break;
         case PLAYERFILLED:
         case CROSS:
            bitmap = Image::Images[ IMAGE_CROSS];
            dest.w = bitmap->w;
            dest.h = bitmap->h;            
            SDL_BlitSurface( Image::Images[ IMAGE_SPACE ], 0, board, &dest );         
            break;
         case PLAYER:         
            bitmap = Image::Images[ IMAGE_PLAYER];
            dest.w = bitmap->w;
            dest.h = bitmap->h;
            SDL_BlitSurface( Image::Images[ IMAGE_SPACE ], 0, board, &dest );         
            break;
         case BOX:
            bitmap = Image::Images[ IMAGE_BOX];
            dest.w = bitmap->w;
            dest.h = bitmap->h;            
            SDL_BlitSurface( Image::Images[ IMAGE_SPACE ], 0, board, &dest );
            break;
         case BOXFILLED:
            bitmap = Image::Images[ IMAGE_BOXFILLED ];
            dest.w = bitmap->w;
            dest.h = bitmap->h;            
            SDL_BlitSurface( Image::Images[ IMAGE_SPACE ], 0, board, &dest );
            
            break;
         }                                    
         SDL_BlitSurface( bitmap, 0, board, &dest );                 
      }
   }   

   dest.x = (width-MapLengthX*20) >> 1;
   dest.y = (height-MapLengthY*20) >> 1;
   dest.w = MapLengthX*20;
   dest.h = MapLengthY*20;

   SDL_BlitSurface( board, 0, screen, &dest );
   return TRUE;
}

int App::DrawText()
{
   SDL_Color white = { 150, 150, 150, 0 };
   SDL_Color red = { 255, 0, 0 , 0 };
   SDL_Rect dest, dest2;      
   char buffer[100];   

   sprintf( buffer, "Level #%i     Pushes: %i", level, pushes );

   SDL_Surface *ttf        = TTF_RenderText_Solid(fntMain, buffer, white);   
   SDL_Surface *ttfSokoban = TTF_RenderText_Solid(fntMain, "sdlSokoban", red);   
   SDL_Surface *ttfRyan    = TTF_RenderText_Solid(fntMain, "By: Ryan Phillips", white);   

   if( ttf == NULL )
      return FALSE; 

   dest.x = 0;
   dest.y = height - 15;
   dest.w = width;//ttf->w;
   dest.h = height;//ttf->h;

   dest2.x = 5;
   dest2.y = 0;
   dest2.w = ttfSokoban->w;
   dest2.h = ttfSokoban->h;
   
   SDL_FillRect( status, 0, SDL_MapRGB( screen->format, 0,0,0 ) );
   SDL_BlitSurface( ttfSokoban, 0, status, &dest2 );

   dest2.x = 5 + ttfSokoban->w + 20;
   dest2.y = 0;
   dest2.w = ttfRyan->w;
   dest2.h = ttfRyan->h;
   
   SDL_BlitSurface( ttfRyan, 0, status, &dest2 );


   dest2.x = (ttf->w + width - MapLengthX*20 ) >> 1;
   dest2.y = 0;
   dest2.w = ttf->w;
   dest2.h = ttf->h;
   
   SDL_BlitSurface( ttf, 0, status, &dest2 );


   SDL_BlitSurface( status, 0, screen, &dest );
   SDL_UpdateRect(  screen, dest.x,dest.y,dest.w,dest.h );

   return TRUE;
}

int App::ResetGame()
{
   SDL_FreeSurface( board );  
   MapStore.ReadMap( level );
   changed_map = TRUE;
   MapLengthX = MapStore.GetLimitX();
   MapLengthY = MapStore.GetLimitY();
   board = SDL_CreateRGBSurface (screen->flags, MapLengthX*20, MapLengthY*20, bpp, 0, 0, 0, 0);
   SetDirty( TRUE );   
   pushes= 0;
   Draw();
   return TRUE;
}

int App::DrawBorder()
{   
/*   SDL_Rect dest;      
   int Width = 400;
   int Height = 400;
     
   SDL_SetColorKey( board, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(screen->format, 0,0,0) );

   dest.x = width*.05;
   dest.y = height*.05;
   dest.w = width;
   dest.h = height;
 
   lock_graphics( board );
   
	
   unlock_graphics( board );
   SDL_BlitSurface( board, 0, screen, &dest );*/
   return TRUE;
}