/***************************************************************************
                          GFX.cpp  -  Graphics handling for SDL.
                             -------------------
    begin                : Sun Dec 10 2000    
    email                : ryan@trolocsis.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/* 
 * SDLRoids - An Astroids clone.
 * 
 * Copyright (c) 2000 David Hedbor <david@hedbor.org>
 * 	based on xhyperoid by Russel Marks.
 * 	xhyperoid is based on a Win16 game, Hyperoid by Edward Hutchins 
 *
 *    Note from Ryan Phillips:
 *      Some of this code was ripped and/or modified from xhyperoid.
 * 
 */

#include <string>
#include <math.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "bool.h"
#include "GFX.h"

extern SDL_Surface* screen;
 
int IsKeyDown(int key)
{
  Uint8 *keystate;
  keystate = SDL_GetKeyState(NULL);
  switch(key)
  {
  case KEY_F1:		return keystate[SDLK_F1];
  case KEY_TAB:	return keystate[SDLK_TAB]; 
  case KEY_S:		return keystate[SDLK_s]; 
  case KEY_LEFT:	return keystate[SDLK_LEFT]; 
  case KEY_RIGHT:	return keystate[SDLK_RIGHT]; 
  case KEY_DOWN:	return keystate[SDLK_DOWN]; 
  case KEY_UP:		return keystate[SDLK_UP]; 
  case KEY_SPACE:	return keystate[SDLK_SPACE]; 
  case KEY_ESC:	return keystate[SDLK_ESCAPE]; 
  default:
    return 0;
  }
}

void fast_putpixel1( SDL_Surface* buffer, Uint16 x, Uint16 y, Uint32 pixel)
{
  *((Uint8 *)buffer->pixels + y * buffer->pitch + x) = pixel;
}
void fast_putpixel2( SDL_Surface* buffer, Uint16 x, Uint16 y, Uint32 pixel)
{
  *((Uint16 *)buffer->pixels + y * buffer->pitch/2 + x) = pixel;
}

void fast_putpixel3( SDL_Surface* buffer, Uint16 x, Uint16 y, Uint32 pixel) 
{ 
    Uint8 *pix = (Uint8 *)buffer->pixels + y * buffer->pitch + x * 3; 
    if(SDL_BYTEORDER == SDL_BIG_ENDIAN) { 
        pix[2] = pixel; 
        pix[1] = pixel >> 8; 
        pix[0] = pixel >> 16; 
    } else { 
        pix[0] = pixel; 
        pix[1] = pixel >> 8; 
        pix[2] = pixel >> 16; 
    } 
} 
void fast_putpixel4( SDL_Surface* buffer, Uint16 x, Uint16 y, Uint32 pixel)
{
  *((Uint32 *)buffer->pixels + y * buffer->pitch/4 + x) = pixel;
}
void (*fast_putpixel)( SDL_Surface* buffer, Uint16 x, Uint16 y, Uint32 pixel);

inline static void _sdl_drawpixel( SDL_Surface* buffer, int x, int y, int r, int g, int b)
{
  Uint8 *bits, bpp;
  Uint32 pixel = SDL_MapRGB(buffer->format, r, g, b);
  //  printf("Drawing %dx%d in %d,%d,%d\n", x, y, r, g, b);
  bpp = buffer->format->BytesPerPixel;
  bits = ((Uint8 *)buffer->pixels)+y*buffer->pitch+x*bpp;

  /* Set the pixel */
  switch(bpp)
  {
   case 1:
    *((Uint8 *)(bits)) = (Uint8)pixel;
    break;
   case 2:
    *((Uint16 *)(bits)) = (Uint16)pixel;
    break;
   case 3: { /* Format/endian independent */
     Uint8 r, g, b;

     r = (pixel>>buffer->format->Rshift)&0xFF;
     g = (pixel>>buffer->format->Gshift)&0xFF;
     b = (pixel>>buffer->format->Bshift)&0xFF;
     *((bits)+buffer->format->Rshift/8) = r; 
     *((bits)+buffer->format->Gshift/8) = g;
     *((bits)+buffer->format->Bshift/8) = b;
   }
   break;
   case 4:
    *((Uint32 *)(bits)) = (Uint32)pixel;
    break;
  }
}

static void sdl_drawpixel(SDL_Surface* buffer, int x, int y, Uint32 color)
{
  if(x >= buffer->w || x < 0 ||
     y >= buffer->h || y < 0) {
    return;
  } 

  (*fast_putpixel)( buffer, x, y, color);
}

/* Draw a circle in the current color. Based off the version in SGE. */
void drawcircle(SDL_Surface* buffer, Sint16 x, Sint16 y, Sint32 r, Uint32 color)
{
  Sint32 cx = 0;
  Sint32 cy = r;
  Sint32 df = 1 - r;
  Sint32 d_e = 3;
  Sint32 d_se = -2 * r + 5;
  do {
    sdl_drawpixel( buffer, x+cx, y+cy, color);
    sdl_drawpixel( buffer, x-cx, y+cy, color);
    sdl_drawpixel( buffer, x+cx, y-cy, color);
    sdl_drawpixel( buffer, x-cx, y-cy, color);
    sdl_drawpixel( buffer, x+cy, y+cx, color);
    sdl_drawpixel( buffer, x+cy, y-cx, color);
    sdl_drawpixel( buffer, x-cy, y+cx, color);
    sdl_drawpixel( buffer, x-cy, y-cx, color);
    
    if (df < 0)  {
      df   += d_e;
      d_e  += 2;
      d_se += 2;
    } else {
      df   += d_se;
      d_e  += 2;
      d_se += 4;
      cy--;
    }
    cx++;
  } while(cx <= cy);
}


/* Draw an horizontal line in the current color */
void draw_horzline( SDL_Surface* buffer, Sint16 x1, Sint16 x2, Sint32 y, Uint32 color )
{
  int i;
  if (x1 < x2) {
    for (i = x1; i <= x2; i++)
      sdl_drawpixel( buffer, i, y, color);
  } else {
    for (i = x2; i <= x1; i++)
      sdl_drawpixel( buffer, i, y, color);
  }  
  return;
}

/* Draw an vertical line in the current color */
void draw_vertline( SDL_Surface* buffer, Sint16 x, Sint16 y1, Sint32 y2, Uint32 color)
{
  int i;
  if (y1 < y2) {
    for (i = y1; i <= y2; i++)
      sdl_drawpixel(buffer, x, i, color);
  } else {
    for (i = y2; i <= y1; i++)
      sdl_drawpixel(buffer, x, i, color);
  }  
  return;
}

/* Draw a line between two coordinates */
void drawline( SDL_Surface* buffer, int x1,int y1,int x2,int y2, Uint32 color)
{
  int d, x, y, ax, ay, sx, sy, dx, dy;
  if((dx = x2 - x1) == 0) { /* vertical line */
    draw_vertline(buffer, x1, y1, y2, color);
    return;
  }
  if((dy = y2 - y1) == 0) { /* horizontal line */
    draw_horzline(buffer, x1, x2, y1, color);
    return;
  }
  ax = ABS(dx)<<1;  sx = SGN(dx);
  ay = ABS(dy)<<1;  sy = SGN(dy);

  x = x1;
  y = y1;
  if (ax>ay)
  {		/* x dominant */
    d = ay-(ax>>1);
    for(;;)
    {
      sdl_drawpixel(buffer, x, y, color);
      if (x==x2) return;
      if (d>=0)
      {
	y += sy;
	d -= ax;
      }
      x += sx;
      d += ay;
    } 
  } else {			/* y dominant */
    d = ax-(ay>>1);
    for (;;) 
    {
      sdl_drawpixel(buffer, x,y, color);
      if (y==y2) return;
      if (d>=0)
      {
	x += sx;
	d -= ay;
      }
      y += sy;
      d += ax;
    }
  }
}


void unlock_graphics( SDL_Surface* buffer )
{
  if ( SDL_MUSTLOCK(buffer) ) 
  {
    SDL_UnlockSurface(buffer);
  }
}
void lock_graphics( SDL_Surface* buffer) 
{
  if ( SDL_MUSTLOCK(buffer) ) 
  {
    if ( SDL_LockSurface(buffer) < 0 ) 
    {
      return;
    }
  }
}


SDL_Surface* bmp2surface( char* bitmapName )
{
  int i;
  SDL_Surface *loaded, *converted;
  
  if( (loaded = IMG_Load(bitmapName)) ) 
  {
    /* Ugly hack to set the "transparent" color to black :-) */
    if(loaded->format->BytesPerPixel == 1) {
      for(i = 0; i < loaded->format->palette->ncolors; i++)
      {
	      /* Make the "transparent" color black... */
	      if(loaded->format->palette->colors[i].r == 0 &&
	         loaded->format->palette->colors[i].g == 0   &&
	         loaded->format->palette->colors[i].b == 0) 
         {
	         loaded->format->palette->colors[i].r =
	         loaded->format->palette->colors[i].g = 
	         loaded->format->palette->colors[i].b = 0;
	         break;
         }
      }
    }
    converted = SDL_DisplayFormat(loaded);
    SDL_FreeSurface(loaded);
    SDL_SetColorKey(converted, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(screen->format, 0,0,0) );
    return converted;    
  } else {    
    return NULL;
  }
}

SDL_Surface* init_graphics( int width, int height, int bpp )
{
  SDL_Surface* buffer;
  int video_flags;
  video_flags = SDL_SWSURFACE;
  //video_flags |= SDL_ASYNCBLIT;
  video_flags |= SDL_HWSURFACE; // |SDL_FULLSCREEN;
  
  if ( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO) < 0 ) {
    fprintf(stderr,
	    "Couldn't initialize SDL: %s\n", SDL_GetError());
    exit(1);
  }
  
  SDL_WM_SetCaption("SDL Sokoban", NULL);
  
  atexit(SDL_Quit);
  SDL_ShowCursor(0);

  /* Initialize the display */
  buffer = SDL_SetVideoMode(width, height, bpp, video_flags);
  if ( buffer == NULL ) {
     return NULL;
  }

  if( TTF_Init() < 0 ) {
     return NULL;
  }
  
  /* Decide which pixel drawing function to use. This is an optimization */
  switch (buffer->format->BytesPerPixel) {
   case 1:
    fast_putpixel = fast_putpixel1;
    break;
   case 2:
    fast_putpixel = fast_putpixel2;
    break;
   case 3:
    fast_putpixel = fast_putpixel3;
    break;
   case 4:
    fast_putpixel = fast_putpixel4;
    break;
   default:
    fprintf(stderr, "Unknown video bytes-per-pixel!\n");
    exit(1);
  }
  return buffer;   
}

void SetCaptions( char* Name, char* Icon )
{
   SDL_WM_SetCaption( Name, Icon);
   //SDL_WM_SetIcon(bitmaps[BMP_ICON], NULL);
}

void poll_keyboard()
{
  SDL_Event event;
  while( SDL_PollEvent(&event) ) 
  {
    switch (event.type) 
    {
      case SDL_QUIT:
         SDL_Quit();
         exit(0);
      case SDL_KEYDOWN:
         if(event.key.keysym.sym == SDLK_RETURN) 
         {
         }
         break;
      case SDL_KEYUP:
         if(event.key.keysym.sym == SDLK_DOWN ||
   	      event.key.keysym.sym == SDLK_UP)
         {
         }
         break;
    }
  }
}

void exit_graphics(void)
{

}




