/***************************************************************************
                          Image.cpp  -  A list of bitmaps
                             -------------------

    This implementation uses the following map to store surfaces.
        map< std::string, SDL_Surface* > Image::Images;
    So to access the background image all you have to do is call:
        Image::Images["BG"]
    This will return the surface that the background image is currently loaded on.

    begin                : Sun Dec 10 2000
    copyright            : (C) 2000 by Ryan Phillips
    email                : ryan@trolocsis.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <iostream>
#include <string>
#include <map>
#include <algorithm>
#include "Image.h"
#include "GFX.h"
#include "bool.h"

// Initiate the static variable found in the Image Class
std::map< std::string, SDL_Surface* > Image::Images;

// Remove and free all the images
template <class FI, class FIB>
void map_delete(FI first, FIB last) {
    while (first != last)
    {
        SDL_FreeSurface( (*first++).second );      
    }
}


int Image::AddBitmap( char* storeName,  char* filename )
{
   SDL_Surface* loaded = bmp2surface( filename );
   if( loaded == 0 )
      return FALSE;

   Images[ storeName ] = loaded;
   
   return TRUE;
}

void Image::FreeImages()
{
   map_delete( Images.begin(), Images.end() );
}

Image::Image()
{
   FreeImages();
}

Image::~Image()
{
   FreeImages();
}